/*	$CoreSDI: srpp.c,v 1.2 2001/08/13 20:18:56 claudio Exp $	*/

/*
 *   Copyright (c) 2000, 2001, Core SDI S.A., Argentina
 *   All rights reserved
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *  
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *   3. Neither name of the Core SDI S.A. nor the names of its contributors
 *      may be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 *  
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 *   IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *   OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *   IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *   NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *   THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *   THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * This program prints on standard output the SRPPass resource data.
 * Author: Claudio Castiglia <claudio@core-sdi.com>
 *
 * Compiles with:
 *   cc srpp.c -o srpp -I../../../include -L../../../lib -lcommon -lcrypto
 */

#include <sys/param.h>
#include <sys/uio.h>

#include <ctype.h>
#include <err.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>
#include <openssl/bn.h>
#include <openssl/err.h>
#include <openssl/evp.h>

#ifdef __OpenBSD__
#include <sha1.h>
#else
#include <openssl/sha.h>
#define SHA1_DIGESTSIZE		SHA_DIGEST_LENGTH
#define SHA1_CTX		SHA_CTX
#define SHA1Init		SHA1_Init
#define SHA1Update		SHA1_Update
#define SHA1Final		SHA1_Final
#endif

#include "sysdep.h"
#include "packet.h"
#include "resource.h"
#include "modtypes.h"
#include "commands.h"
#include "authargs.h"
#include "log.h"

#define AUTH_SRP		"auth srp: "

#ifndef SRP_PASS_RNAME
#define SRP_PASS_RNAME		"SRPPass"	/* Verifier n salt resource */
#endif

#define NBITS			2048
#define NBYTES			(NBITS >> 3)
#define ERR_STRLEN		121		/* See ERR_error_string(3) */

static const unsigned char prime[] = {
	0xFD, 0x66, 0x54, 0x6A, 0xF0, 0x8A, 0x2A, 0x30,
	0x4B, 0x46, 0xB8, 0x9D, 0x7B, 0x70, 0xA4, 0x2D,
	0x4C, 0x57, 0xB3, 0xE9, 0x1B, 0xC6, 0xD2, 0x9C,
	0x9F, 0x74, 0x16, 0x43, 0x35, 0x36, 0x84, 0xE0,
	0x92, 0xE3, 0x25, 0x8B, 0xD9, 0x72, 0x18, 0x7B,
	0x1F, 0x23, 0xC8, 0xC2, 0x19, 0xFC, 0x27, 0xCE,
	0x1B, 0x1C, 0x2C, 0xCB, 0x40, 0xCF, 0x2C, 0xEF,
	0x4B, 0x89, 0x69, 0x9F, 0xBD, 0x02, 0x35, 0x72,
	0xA8, 0xC7, 0x80, 0xAC, 0xA4, 0x59, 0xDC, 0x95,
	0x86, 0xC6, 0xE8, 0x53, 0x38, 0x6A, 0xB5, 0x80,
	0x79, 0x83, 0x91, 0x79, 0x5E, 0xF3, 0xD6, 0x67,
	0x65, 0xAF, 0x06, 0xEA, 0x8F, 0xBB, 0x3D, 0xF9,
	0x91, 0xE1, 0x4A, 0xE2, 0x59, 0x11, 0xDD, 0xC7,
	0xA4, 0xE8, 0x10, 0x17, 0x2F, 0xFD, 0xCA, 0xCA,
	0x32, 0x70, 0x29, 0xE9, 0xDA, 0xD6, 0x54, 0x3F,
	0x03, 0x10, 0xEE, 0x73, 0x36, 0x75, 0xA2, 0xB1,
	0x1D, 0xCD, 0x63, 0xFF, 0x0E, 0x6D, 0x63, 0x3E,
	0x87, 0xF6, 0xD7, 0x95, 0x36, 0x08, 0x04, 0xC0,
	0x63, 0x4D, 0x49, 0x10, 0x77, 0x50, 0xB3, 0xA9,
	0xAE, 0x98, 0x85, 0x96, 0xA0, 0x43, 0x95, 0x7F,
	0xDE, 0x3E, 0x6B, 0xE1, 0x29, 0x1A, 0xC2, 0x74,
	0xA8, 0x97, 0x91, 0xB5, 0xF0, 0x34, 0xF9, 0xAA,
	0xC3, 0x0A, 0x46, 0x28, 0x23, 0x16, 0xE7, 0x63,
	0x74, 0xAC, 0xA3, 0x4E, 0xA7, 0xF5, 0x74, 0x9E,
	0x19, 0x07, 0xE9, 0xE7, 0xC4, 0x56, 0xC4, 0x7C,
	0x2E, 0x71, 0x19, 0xA3, 0x1B, 0x4D, 0x88, 0x89,
	0x66, 0xD4, 0xE9, 0x53, 0xCA, 0x68, 0x24, 0x4A,
	0xBA, 0x8D, 0x8C, 0x04, 0x31, 0x9F, 0x70, 0x19,
	0xBD, 0xDB, 0x48, 0xF9, 0xF0, 0x42, 0xA8, 0xCE,
	0x65, 0xF9, 0x87, 0xAB, 0x57, 0x3E, 0xE3, 0x61,
	0xFA, 0xBC, 0xD4, 0xC3, 0x48, 0x2D, 0x9D, 0x3F,
	0x70, 0xB9, 0x17, 0x1B, 0xC8, 0x16, 0xB9, 0x3F
};

static const unsigned char generator[] = {
	0xA8, 0xEE, 0xE2, 0xF1, 0xF5, 0xB1, 0x71, 0x75,
	0x87, 0x84, 0x7B, 0x13, 0xA7, 0xA0, 0x6D, 0x73,
	0x88, 0x3A, 0x77, 0xF0, 0xBD, 0x2F, 0x37, 0x13,
	0x14, 0xF8, 0x0E, 0xD7, 0x78, 0xCF, 0x03, 0x40,
	0x61, 0xEC, 0xC3, 0xB2, 0x90, 0xF6, 0xBA, 0xFC,
	0xBF, 0x6D, 0x30, 0x81, 0x66, 0xA8, 0x1A, 0x89,
	0x67, 0x68, 0x1D, 0xDC, 0xD5, 0xDF, 0x73, 0x4A,
	0x32, 0x5B, 0x9B, 0xBF, 0xD3, 0x56, 0xCE, 0x4C,
	0x70, 0x85, 0x00, 0x73, 0x18, 0x3B, 0xE8, 0x63,
	0xAF, 0x2F, 0x45, 0x8C, 0xD0, 0x47, 0x23, 0xAA,
	0xFB, 0xAD, 0x0B, 0xA6, 0x3F, 0x4D, 0x39, 0x9A,
	0x43, 0xCA, 0x04, 0x9C, 0x5F, 0xD2, 0x29, 0x51,
	0x0B, 0xEB, 0x87, 0x41, 0x90, 0xB6, 0x93, 0xDA,
	0x6D, 0xF0, 0x0A, 0xBA, 0x1F, 0xFE, 0x87, 0x31,
	0x76, 0xF5, 0x71, 0x46, 0x91, 0xE4, 0x38, 0x2A,
	0x02, 0x0B, 0x49, 0xA2, 0x24, 0x4E, 0x6C, 0x76,
	0x13, 0xDE, 0x42, 0xAA, 0x09, 0x9E, 0x42, 0x29,
	0xAF, 0xF9, 0xE5, 0x0E, 0x24, 0x05, 0x58, 0x80,
	0x42, 0x33, 0x86, 0x0A, 0xFA, 0x35, 0xCD, 0x1B,
	0xC9, 0xBB, 0x03, 0xB9, 0xC0, 0x2D, 0x0E, 0x55,
	0x3E, 0xD4, 0x47, 0xEB, 0x70, 0xBC, 0x81, 0xA3,
	0x1B, 0x0F, 0xB6, 0x79, 0x4A, 0xCD, 0xFB, 0xC7,
	0x2C, 0xB1, 0x84, 0x1A, 0xC2, 0x0F, 0x44, 0xEC,
	0xF8, 0x73, 0x17, 0x89, 0xC5, 0x4E, 0x4D, 0xBE,
	0xBB, 0x5A, 0x9B, 0xEF, 0xD8, 0x39, 0xD8, 0x52,
	0xC9, 0xA0, 0xBB, 0xC2, 0x12, 0x33, 0xB0, 0x5B,
	0x99, 0xE3, 0x46, 0x37, 0xDC, 0x45, 0x6D, 0x87,
	0x27, 0x09, 0x08, 0x02, 0xCB, 0xBF, 0xA0, 0x11,
	0x29, 0x3C, 0xDB, 0x51, 0x4A, 0xD7, 0x1B, 0x34,
	0x43, 0xFB, 0xAF, 0xC7, 0x8F, 0x7F, 0x42, 0x41,
	0x51, 0xD3, 0x38, 0x82, 0x30, 0x1E, 0x68, 0xD4,
	0xF5, 0xD0, 0xBA, 0x12, 0x85, 0x64, 0x7B, 0x7F
};

static BIGNUM	n;				/* Prime module	*/
static BIGNUM 	g;				/* Generator	*/


/*
 * _fatal_bn():
 *	Log fatal bn (big number) library error and exit(3).
 */
static void
_fatal_bn(const char *cname)
{
	char errb[ERR_STRLEN];

	ERR_error_string(ERR_get_error(), errb);
	fatal(EX_SOFTWARE, "srp: %s: %s.", cname, errb);
}


/*
 * _rand_bn():
 *	Generate a random big number (!=0 && !=1) modulo n.
 */
static void
_rand_bn(BIGNUM *bn)
{
	do {
		if (!BN_rand(bn, NBITS, 0, 0))
			_fatal_bn("_rand_bn()");
	} while (BN_is_zero(bn) || BN_is_one(bn));
}


/*
 * _init()
 *	Initialization function, called once by dlopen(3).
 */
void
_init()
{
	BN_init(&n);
	BN_init(&g);
	if (BN_bin2bn(prime, sizeof(prime), &n) == NULL ||
	    BN_bin2bn(generator, sizeof(generator), &g) == NULL)
		_fatal_bn("_init()");
}


/*
 * _fini()
 *	Finalization function, called once by last dlclose(3).
 */
void
_fini()
{
	BN_free(&n);
	BN_free(&g);
}


/*
 * filter_resource()
 *	Detect a SRP_PASS_RNAME resource and code its correct value;
 *	return 0 on success and -1 on error and set errno.
 *	On fatal errors program is terminated.
 */
static int
_filter_resource(AUTHCON *ct, struct autharg_filter *args)
{
	SHA1_CTX	sha1_ctx;
	BN_CTX		bnctx;
	BIGNUM		salt, verifier, x;
	unsigned char	bsalt[NBYTES], xdigest[SHA1_DIGESTSIZE];

	if (args->rname == NULL || args->value == NULL || !args->vsize) {
		errno = EINVAL;
		return (-1);
	}

	/* Just filter srp passwords */
	if (!strcmp(args->rname, SRP_PASS_RNAME)) {

		BN_CTX_init(&bnctx);
		BN_init(&salt);
		BN_init(&x);
		BN_init(&verifier);

		/*
		 * Calculate:
		 *	salt = rand() % n
		 *	x = HASH(salt, password)
		 *	verifier = g exp x
		 */
		_rand_bn(&salt);
		BN_bn2bin(&salt, bsalt);
		SHA1Init(&sha1_ctx);
		SHA1Update(&sha1_ctx, bsalt, sizeof(bsalt));
		SHA1Update(&sha1_ctx, args->value, args->vsize);
		SHA1Final(xdigest, &sha1_ctx);
		if (BN_bin2bn(xdigest, sizeof(xdigest), &x) == NULL ||
		    !BN_mod_exp(&verifier, &g, &x, &n, &bnctx))
			_fatal_bn("_filter_resource()");

		/* Get memory for the filtered value */
		args->fvsize = NBYTES + BN_num_bytes(&verifier);
		args->fvalue = malloc(args->fvsize);
		if (args->fvalue != NULL) {
			/*
			 * The resource value is the binary form of
			 * both salt and verifier data concatenated.
			 */
			memcpy(args->fvalue, bsalt, sizeof(bsalt));
			BN_bn2bin(&verifier, args->fvalue + sizeof(bsalt));
		}

		BN_clear_free(&verifier);
		BN_clear_free(&x);
		BN_clear_free(&salt);
		BN_CTX_free(&bnctx);

		if (args->fvalue == NULL)
			return (-1);
	}

	return (0);
}


/*
 * main()
 */
int
main(int argc, char **argv)
{
	extern char		*__progname;
	struct autharg_filter	 f;
	int			 i, j;

	if (argc < 2) {
		fprintf(stderr, "Usage: %s <password>\n", __progname);
		exit(-1);
	}

	_init();
	f.rname = SRP_PASS_RNAME;
	f.value = argv[1];
	f.vsize = strlen(argv[1]);
	if (_filter_resource(NULL, &f) < 0)
		err(-1, "Can't filter resource");
	_fini();

	printf("[ " SRP_PASS_RNAME " ]");
	for (i = 0; i < f.fvsize; i += 16) {
		printf("\n\t");
		for (j = 0; j < 16; j++)
			printf("\\%02X", *((unsigned char *) f.fvalue + j + i));
	}
	
	printf("\n");
	free(f.fvalue);
	return (0);
}
